/*
 * Mainboart_UART_functions.c
 *
 * Created: 21.01.2014 10:18:41
 *  Author: Alex
 */ 

#include "Mainboard_UART_functions.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <stdio.h> 
#define F_CPU 20E6
#include <util/delay.h>
#include "uart.h"
#include "LTC2486_functions.h"
#include "sbopcodes_custom.h"

void ADC_handler(uint8_t adcval, uint8_t triggerlevel, bool *adctoggle){
	if ((adcval < triggerlevel)&&(*adctoggle)){								// If ADC is enabled and Trigger-Signal < Threshold: Send Line
		char buffer[40];
		sprintf( buffer, "#SSUT\r\n");
		uart_puts( buffer );
		*adctoggle = 0;
	}
	else if ((adcval > triggerlevel)&&(!*adctoggle)){
		char buffer[40];
		sprintf( buffer, "#SSOT\r\n");
		uart_puts( buffer );
		*adctoggle = 1;
	}
}

 void ADC_init(){
	 
	 ADMUX |= (1<<REFS0)|(1<<REFS1);
	 //ADMUX |= (1<<REFS0)|(1<<ADLAR);		// AVCC Reference with external capacitor at AREF; Left Adjust result (8MSBits from ADCH); ADC0 select
	 ADCSRA |= (1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0)|(1<<ADATE);			// 1/ 128Prescaling bit for ~150kHz ADC clock, Single Ended Conversion; Auto Trigger Enable, Free Running Mode		//
	 ADCSRA |= (1<<ADEN)|(1<<ADSC);												// Enable ADC, Set ADC Start Conversion Bit
 }
 
 void ADC_disable(){
	 ADCSRA &= ~(1<<ADEN);			// Disable ADC for Power Saving
 }
 
 void ADC_enable(){
	 ADCSRA |= (1<<ADEN);			// Enable ADC
 }
 
 uint16_t ADC_gettwobytevalue(){
	 uint16_t val =0;
	 uint8_t lower;
	 uint8_t upper;
	 
	 lower = ADCL;
	 upper = ADCH;
	 val = ((uint16_t)upper <<8)|lower;
	 
	 return val;
 }
 
 uint8_t ADC_get_value(){
	 uint8_t val =0;
	 val = ADCH;
	 return val;
 }

void LED2_on() {
	PORTC |= (1<<PINC6);
}

void LED2_off() {
	PORTC &= ~(1<<PINC6);
}

void Bluetooth_Reset(){				// Reset Bluetooth-Module
	PORTC &= ~(1<<PINC3);
	_delay_ms(200);
	PORTC |= (1<<PINC3);
	_delay_ms(200);	
}

void channel_administrator (conf_stat *conf, meas_info *info){
	info->active_ch = conf->active_ch;			// Save last configuration for transmission of value that we will get at next SPI cycle
	info->active_mod = conf->active_mod;
	info->lambda = conf->active_lambda;
	info->speedmode = conf->speedmode;
	info->temp = conf->temp_flag;
	
	if ((conf->conf_list[conf->conf_cnter]== 'X')&&(conf->active_lambda==true))		// End of List reached?
	conf->start_flag=true;
		
	if (conf->start_flag){				// Start anew
		conf->active_lambda = false;
		conf->conf_cnter=0;
		
		switch(conf->conf_list[conf->conf_cnter]){
			case '0':{
				conf->active_ch=0;
				conf->active_mod=0;
				break;
			}
			case '1':{
				conf->active_ch=1;
				conf->active_mod=0;
				break;
			}
			case '2':{
				conf->active_ch=2;
				conf->active_mod=0;
				break;
			}
			case '3':{
				conf->active_ch=3;
				conf->active_mod=0;
				break;
			}
			case '4':{
				conf->active_ch=0;
				conf->active_mod=1;
				break;
			}
			case '5':{
				conf->active_ch=1;
				conf->active_mod=1;
				break;
			}
			case '6':{
				conf->active_ch=2;
				conf->active_mod=1;
				break;
			}
			case '7':{
				conf->active_ch=3;
				conf->active_mod=1;
				break;
			}
			case '8':{
				conf->active_ch=0;
				conf->active_mod=2;
				break;
			}
			case '9':{
				conf->active_ch=1;
				conf->active_mod=2;
				break;
			}
			case 'A':{
				conf->active_ch=2;
				conf->active_mod=2;
				break;
			}
			case 'B':{
				conf->active_ch=3;
				conf->active_mod=2;
				break;
			}
			case 'C':{
				conf->active_ch=0;
				conf->active_mod=3;
				break;
			}
			case 'D':{
				conf->active_ch=1;
				conf->active_mod=3;
				break;
			}
			case 'E':{
				conf->active_ch=2;
				conf->active_mod=3;
				break;
			}
			case 'F':{
				conf->active_ch=3;
				conf->active_mod=3;
				break;
			}
		}
		conf->conf_cnter++;
		conf->start_flag=false;
	}
	
	else if ((conf->start_flag==false)&&(conf->active_lambda==false)){			// If first lamda has been measured: set same channel for second lambda
				conf->active_lambda= true;
	}
			
	else if ((conf->start_flag==false)&&(conf->active_lambda==true)){										// Both wavelengths of one channel have been measured -> next channel
		conf->active_lambda=false;
		
		switch(conf->conf_list[conf->conf_cnter]){
			case '0':{
				conf->active_ch=0;
				conf->active_mod=0;
				conf->conf_cnter++;
				break;
			}
			case '1':{
				conf->active_ch=1;
				conf->active_mod=0;
				conf->conf_cnter++;
				break;
			}
			case '2':{
				conf->active_ch=2;
				conf->active_mod=0;
				conf->conf_cnter++;
				break;
			}
			case '3':{
				conf->active_ch=3;
				conf->active_mod=0;
				conf->conf_cnter++;
				break;
			}
			case '4':{
				conf->active_ch=0;
				conf->active_mod=1;
				conf->conf_cnter++;
				break;
			}
			case '5':{
				conf->active_ch=1;
				conf->active_mod=1;
				conf->conf_cnter++;
				break;
			}
			case '6':{
				conf->active_ch=2;
				conf->active_mod=1;
				conf->conf_cnter++;
				break;
			}
			case '7':{
				conf->active_ch=3;
				conf->active_mod=1;
				conf->conf_cnter++;
				break;
			}
			case '8':{
				conf->active_ch=0;
				conf->active_mod=2;
				conf->conf_cnter++;
				break;
			}
			case '9':{
				conf->active_ch=1;
				conf->active_mod=2;
				conf->conf_cnter++;
				break;
			}
			case 'A':{
				conf->active_ch=2;
				conf->active_mod=2;
				conf->conf_cnter++;
				break;
			}
			case 'B':{
				conf->active_ch=3;
				conf->active_mod=2;
				conf->conf_cnter++;
				break;
			}
			case 'C':{
				conf->active_ch=0;
				conf->active_mod=3;
				conf->conf_cnter++;
				break;
			}
			case 'D':{
				conf->active_ch=1;
				conf->active_mod=3;
				conf->conf_cnter++;
				break;
			}
			case 'E':{
				conf->active_ch=2;
				conf->active_mod=3;
				conf->conf_cnter++;
				break;
			}
			case 'F':{
				conf->active_ch=3;
				conf->active_mod=3;
				conf->conf_cnter++;
				break;
			}
		}
	}
}

void NIRS_module_reset(uint8_t module){
	switch (module){
		case 1:{
			PORTD |= (1<<PD5);
			_delay_ms(1);
			PORTD &= ~(1<<PD5);
			break;
		}
		case 2:{
			PORTD |= (1<<PD4);
			_delay_ms(1);
			PORTD &= ~(1<<PD4);
			break;
		}
		case 3:{
			PORTD |= (1<<PD6);
			_delay_ms(1);
			PORTD &= ~(1<<PD6);
			break;
		}
		case 0:{
			PORTD |= (1<<PD7);
			_delay_ms(1);
			PORTD &= ~(1<<PD7);
			break;
		}
	}
}

void NIRS_module_trigger (uint8_t module){
	switch (module){
		case 1:{
			PORTA |= (1<<PA3);			
			break;
		}
		case 2:{
			PORTC |= (1<<PC7);
			break;
		}
		case 3:{
			PORTA |= (1<<PA6);
			break;
		}
		case 0:{
			PORTB |= (1<<PB2);
			break;
		}
	}
}

void NIRS_module_deactivate(uint8_t module){
	switch (module){
		case 1:{
			PORTA &= ~(1<<PA3);
			break;
		}
		case 2:{
			PORTC &= ~(1<<PC7);
			break;
		}
		case 3:{
			PORTA &= ~(1<<PA6);
			break;
		}
		case 0:{
			PORTB &= ~(1<<PB2);
			break;
		}
	}
}

void NIRS_module_set_channel (uint8_t module, uint8_t channel){
	// Clear NIRS control outputs
	PORTA &= 0x01;	
	PORTB &= 0xF0;
	PORTC &= 0x7F;
	PORTD &= 0x0F;
	
	switch (module){
		case 1:{
			switch (channel){
				case 0:{
					break;
				}
				case 1:{
					PORTA |= (1<<PA2);
					break;
				}
				case 2:{
					PORTA |= (1<<PA4);
					break;
				}
				case 3:{
					PORTA |= ((1<<PA4)|(1<<PA2));
					break;
				}
			}
			break;
		}
		case 2:{
			switch (channel){
				case 0:{
					break;
				}
				case 1:{
					PORTB = (1<<PB0);
					break;
				}
				case 2:{
					PORTA = (1<<PA1);
					break;
				}
				case 3:{
					PORTB = (1<<PB0);
					PORTA = (1<<PA1);
					break;
				}
			}
			break;
		}
		case 3:{
			switch (channel){
				case 0:{
					break;
				}
				case 1:{
					PORTA |= (1<<PA5);
					break;
				}
				case 2:{
					PORTA |= (1<<PA7);
					break;
				}
				case 3:{
					PORTA |= (1<<PA5);
					PORTA |= (1<<PA7);
					break;
				}
			}
			break;
		}
		case 0:{
			switch (channel){
				case 0:{
					break;
				}
				case 1:{
					PORTB |= (1<<PB3);
					break;
				}
				case 2:{
					PORTB |= (1<<PB1);
					break;
				}
				case 3:{
					PORTB |= ((1<<PB1)|(1<<PB3));
					break;
				}
			}
			break;
		}
	}
}

void process_rxbuf (unsigned int c, uint16_t *timer_ms_counter, conf_stat *config){
	
	unsigned int d;
	uint8_t switch_buf = c;

	if (config->config_flag == false){
		if (config->bt_flag == false){
			switch (switch_buf) {
				case STX:{
					config->bt_flag = true;	
					break;
				}
				case 'G':{										// "Go": Reset Timer, Start Running Mode
					config->running_flag = true;
					config->start_flag = true;
					config->empty_buf_flag = true;
					char buffer[80];
					sprintf( buffer, "Module;Channel;Wavelength;Speedmode;ADC-Value;Timer-Value;\r\n");
					uart_puts( buffer );
					sprintf( buffer, "[0-3];[0-3];[0:750nm, 1:850nm];[1:on, 0:off];[16Bit HEX, *5/2^16];[16Bit HEX, 1/100s];\r\n");
					uart_puts( buffer );
					Timer_reset(timer_ms_counter);
					Timer_enable_interrupt();
					break;
				}
				case 'S':{										// "Stop", Stop Running Mode
					config->running_flag = false;
					Timer_disable_interrupt();
					NIRS_module_deactivate(0);
					NIRS_module_deactivate(1);
					NIRS_module_deactivate(2);
					NIRS_module_deactivate(3);
					NIRS_module_reset(0);
					NIRS_module_reset(1);
					NIRS_module_reset(2);
					NIRS_module_reset(3);
					ADC_init();									// Enable ADC for "Go" Triggering
					break;
				}
				case 'E':{										// Toggle spEed Mode
					if (config->speedmode == true){
						config->speedmode = false;
						char buffer[40];
						sprintf( buffer, "Speedmode off\r\n");
						uart_puts( buffer );
					}
					else {
						config->speedmode = true;
						char buffer[40];
						sprintf( buffer, "Speedmode on\r\n");
						uart_puts( buffer );
					}
					break;
				}
				case 'C':{										// Configuration Mode
					config->config_flag=true;
					config->conf_cnter=0;
					char buffer[40];
					sprintf( buffer, "CONFIGURATION MODE: Please list channel-numbers (0-F, sent in ASCII) you wish to activate.\r\n Finish with <X> to exit\r\n");
					uart_puts( buffer );
					// Reset Configuration Matrix
					for (int i=0;i<4;i++){
						for (int k=0;k<4;k++){
							config->conf_matrix[i][k] =0;
						}
					}
					break;
				}
				case 'P':{										// Pause (does not stop timer)
					if (config->running_flag ==true)
					config->running_flag = false;
					else
					config->running_flag = true;
					
					d = uart_getc();
					uart_putc(d);
					d = uart_getc();
					uart_putc(d);
					d = uart_getc();
					uart_putc(d);
					break;
				}
				case 'R':{										// Read Config Data - yet to be implemented
					char buffer[100];
					sprintf( buffer, "\r\nCONFIGURATION: \r\n");
					uart_puts( buffer );
					sprintf( buffer, "    Speedmode: %d\r\n", config->speedmode);
					uart_puts( buffer );
					sprintf( buffer, "    Configuration Matrix: \r\n");
					uart_puts( buffer );
					sprintf( buffer, "    M0  M1  M2  M3\r\n");
					uart_puts( buffer );
					sprintf( buffer, "CH0 %d   %d   %d   %d\r\n", config->conf_matrix[0][0], config->conf_matrix[0][1], config->conf_matrix[0][2],config->conf_matrix[0][3]);
					uart_puts( buffer );
					sprintf( buffer, "CH1 %d   %d   %d   %d\r\n", config->conf_matrix[1][0], config->conf_matrix[1][1], config->conf_matrix[1][2],config->conf_matrix[1][3]);
					uart_puts( buffer );
					sprintf( buffer, "CH2 %d   %d   %d   %d\r\n", config->conf_matrix[2][0], config->conf_matrix[2][1], config->conf_matrix[2][2],config->conf_matrix[2][3]);
					uart_puts( buffer );
					sprintf( buffer, "CH3 %d   %d   %d   %d\r\n", config->conf_matrix[3][0], config->conf_matrix[3][1], config->conf_matrix[3][2],config->conf_matrix[3][3]);
					uart_puts( buffer );
					break;
				}
				case 'T':{										// Next Measurement: Temperature Measurement
					config->temp_flag = true;
					break;
				}
				case 'H':{										// Help
					char buffer[200];
					sprintf( buffer, "\r\n\r\nHELP:\r\n G: Go (Start Measurement, resets Timestamp)\r\n S: Stop Measurement\r\n P: Pause (does not reset timer - to continue send 'P'ause again\r\n R: Read Configuration Data \r\n");
					uart_puts( buffer );
					sprintf( buffer, " T: Temperature Measurement\r\n E: Toggle ADC Speed Mode\r\n");
					uart_puts( buffer );
					sprintf( buffer, " C: Starts Configuration Mode. When entered, send corresponding HEX numbers of the channels you want to activate.\r\n");
					uart_puts( buffer );
					sprintf( buffer, "          Finish with 'X'. Channelnumbers:\r\n");
					uart_puts( buffer );
					
					sprintf( buffer, "              M0  M1  M2  M3\r\n");
					uart_puts( buffer );
					sprintf( buffer, "          CH0 0   4   8   C\r\n");
					uart_puts( buffer );
					sprintf( buffer, "          CH1 1   5   9   D\r\n");
					uart_puts( buffer );
					sprintf( buffer, "          CH2 2   6   A   E\r\n");
					uart_puts( buffer );
					sprintf( buffer, "          CH3 3   7   B   F\r\n");
					uart_puts( buffer );
					sprintf( buffer, "          Example: '01238CX' activates all Channels from Module 0, channel 8 from module 2 and channel C from module 3 \r\n");
					uart_puts( buffer );
					break;
				}
			}
		}
		else {
			switch (switch_buf){
				case ETX:{
					config->bt_flag = false;
					break;
				}
				case MODULE_READY:{
					LED2_on();
					break;
				}
				case SPP_INCOMING_LINK_ESTABLISHED:{
					LED2_off();					
					break;
				}
				case SPP_TRANSPARENT_MODE:{
					LED2_on();
					break;
				}
				case SPP_LINK_RELEASED:{
					config->running_flag = false;
					LED2_on();
					break;
				}
				default:{
					break;
				}
			}
			
		}
	}
	// Set Configuration Matrix
	else if (config->config_flag==true){
		if (switch_buf == '0' || switch_buf == '1' || switch_buf == '2' || switch_buf == '3' || switch_buf == '4' || switch_buf == '5' || switch_buf == '6' || switch_buf == '7' || switch_buf == '8' || switch_buf == '9' || switch_buf == 'A' || switch_buf == 'B' || switch_buf == 'C' || switch_buf == 'D' || switch_buf == 'E' || switch_buf == 'F'|| switch_buf == 'X'  ){
			switch (switch_buf){
				case '0':{
					config->conf_matrix[0][0]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case '1':{
					config->conf_matrix[1][0]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case '2':{
					config->conf_matrix[2][0]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case '3':{
					config->conf_matrix[3][0]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case '4':{
					config->conf_matrix[0][1]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case '5':{
					config->conf_matrix[1][1]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case '6':{
					config->conf_matrix[2][1]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case '7':{
					config->conf_matrix[3][1]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case '8':{
					config->conf_matrix[0][2]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case '9':{
					config->conf_matrix[1][2]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case 'A':{
					config->conf_matrix[2][2]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}	
				case 'B':{
					config->conf_matrix[3][2]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case 'C':{
					config->conf_matrix[0][3]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case 'D':{
					config->conf_matrix[1][3]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}
				case 'E':{
					config->conf_matrix[2][3]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}		
				case 'F':{
					config->conf_matrix[3][3]=true;
					config->conf_list[config->conf_cnter]=switch_buf;
					break;
				}		
				case 'X':{
					config->config_flag = false;
					config->conf_list[config->conf_cnter]=switch_buf;
					char buffer[40];
					sprintf( buffer, "\r\nConfiguration finished. Leaving Config-Mode. \r\n");
					uart_puts( buffer );
					break;
				}																																												
			}
			config->conf_cnter++;
		}
		else{
		char buffer[40];
		sprintf( buffer, "\r\nERROR: Input contained wrong character. Leaving Config-Mode. \r\n");
		config->config_flag=false;
		uart_puts( buffer );	
		}
	}
}

uint16_t generate_configuration (conf_stat *config){
	uint16_t config_buffer;
	
	if (config->temp_flag ==true) {
	config_buffer= LTC2486_create_config(0,0,1,1);
	config->temp_flag=false;
	}
	else {
	config_buffer= LTC2486_create_config(config->active_mod, config->speedmode, 0, 1);		
	}
	return config_buffer;
}

void send_ADC_raw(uint32_t adcreceive){
	unsigned int adcreceive_h = (adcreceive>>16);
	unsigned int adcreceive_l = adcreceive;
	char buffer[40];
	sprintf( buffer, "RAW_ADC:%04X%04X\r\n", adcreceive_h, adcreceive_l);
	uart_puts( buffer );
}

void send_ADC_temp (uint16_t temp){
	// Send ASCII TempBlock: [T|E|M|P|:|TEMPVAL(Hex,4ASCII)|CR|LF]		
	char buffer[40];
	sprintf( buffer, "TEMP[/2.5 in K]:%04X;\r\n", temp);
	uart_puts( buffer );
}

void send_ADC_data (uint8_t active_module, uint8_t active_channel, bool lambda, bool Speedmode, uint16_t adc_value, uint16_t timestamp) {
	// Send ASCII DataBlock: [M|#|;|C|#|;|L|#|;|S|0/1|;|ADCVAL(Hex,4ASCII)|;| |TIMERVAL(Hex,4ASCII)|CR|LF]
	 char buffer[40];
	 sprintf( buffer, "M%d;C%d;L%d;S%d;%04X;%04X;\r\n", active_module, active_channel, lambda, (int)Speedmode, adc_value, timestamp );
	 uart_puts( buffer );
}

void Timer_init(uint16_t ctcval) {
	ctcval = 195;					// Output Compare Match with prescaler 1/1024: Every ~10ms
	PRR &= ~(1<<PRTIM1);			// Power Reduction Register: Enable Timer/Counter1
	OCR1A = ctcval;
	TCCR1B |= (1<<CS12)|(1<<CS10)|(1<<WGM12);	// Prescaler 1/1024, Clear Timer on Compare match (CTC), Mode 4
	TIMSK1 |= (1<<OCIE1A);									// Timer/Counter1 Output Compare Interrupt Request
}

void Timer_reset(uint16_t *timer_ms_counter) {
	uint16_t val = 0;
	timer_ms_counter = 0;
	TCNT1 = val;
}

void Timer_disable_interrupt()
{
	TIMSK1 &= ~(1<<OCIE1A);	
}

void Timer_enable_interrupt()
{
	TIMSK1 |= (1<<OCIE1A);
}